package ftp;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.UnknownHostException;
import java.nio.ByteBuffer;
import java.nio.channels.SocketChannel;

import util.PasvResponse;

public class FtpProtocolInterpreter implements Runnable  {
	private FtpPIStatus status = FtpPIStatus.NotConnected;
	private SocketChannel servChan;
	private String servHostname = null;
	private int port = 21;
	private String user, pass;
	private String absoluteFileName;
	private ByteBuffer readBuf, writeBuf;
	
	
	public FtpProtocolInterpreter() {
		readBuf =  ByteBuffer.allocate(2048);
		writeBuf =  ByteBuffer.allocate(2048);	
	}
	
	
	
	 /**
	  * Initialise les attributs de la classe
	  * @param servHostname: nom de la machine où chaîne contenant l'adresse IP où tourne le serveur FTP
	  * @param port: port utilisé par le serveur
	  * @param user: nom de l'identifiant utilisé pour se connecter au serveur FTP
	  * @param pass: password de l'identifiant.
	  * @param absoluteFileName: nom absolu du fichier à télécharger
	  */
	public FtpProtocolInterpreter(String servHostname, int port, String user, String pass, String absoluteFileName) {
		this();
		this.servHostname = servHostname;
		this.port = port;
		this.user = user;
		this.pass = pass;
		this.absoluteFileName = absoluteFileName;
		
	}
	
	/**
	 * Connecte le PI (Protocole Interpreter) du client FTP au PI du serveur FTP 
	 * @param servHostname
	 * @param port
	 * @throws UnknownHostException
	 * @throws IOException
	 */
	public void connectTo(String servHostname, int port) throws UnknownHostException, IOException {
		if (status != FtpPIStatus.NotConnected) {
			System.out.println("Allready connected. Keep the connexion with " + servHostname);
		} else {

			servChan = SocketChannel.open(new InetSocketAddress(InetAddress.getByName(servHostname), port));
			System.out.println("Connected to " + servHostname);
			this.servHostname = servHostname;
			this.port = port;
			String response = recvStringFromServer();
			if (!cmdOk(response)) {
				servChan.close();
			}
			else {
				status = FtpPIStatus.Connected;
			}

		}
	}
	
	

	/**
	 * Ferme le canal du PI du client FTP.
	 * @throws IOException
	 */
	public void close() throws IOException {
		if (status != FtpPIStatus.NotConnected) {
			servChan.close();
			status = FtpPIStatus.NotConnected;
		}
	}
	
	
	/**
	 * Envoie une ligne de protocole au PI du serveur FTP
	 * @param cmd
	 * @throws IOException
	 */
	private void sendStringToServer(String cmd)  throws IOException {
		writeBuf.clear();
		writeBuf.put(cmd.getBytes());
		writeBuf.flip();
		
		servChan.write(writeBuf);	
		System.out.print("--> " + cmd);
	}
	
	
	/**
	 * Reçoit une ligne de protocole du PI du serveur FTP
	 * @return une chaîne contenant la ligne reçue
	 * @throws IOException
	 */
	private String recvStringFromServer() throws IOException {
		
		readBuf.clear();
		int readSize = servChan.read(readBuf);
		if(readSize == -1) {
			return null;
		}
		
		readBuf.flip();
		byte [] allBytes = new byte[readBuf.limit()];
		readBuf.get(allBytes);
		String response = new String(allBytes);		
		System.out.print(response);
		return response;
	}
	
	
	/**
	 * Analyse du code de la réponse:
	 * les messages commençant par les codes 1xx, 2xx ou 3xx sont des succés.
	 * les messages commançant par les codes 4xx ou 5xx sont des échecs
	 * @param response
	 * @return true en cas succés, false en cas d'échec
	 */
	private static boolean cmdOk(String response) {
		switch (response.charAt(0)) {
		case '1':
		case '2':
		case '3':
			return true;
		case '4':
		case '5':
			return false;
		}
		return false;
	}
	
	
	/**
	 * Logue un utilisateur au serveur FTP.
	 * Prérequis: le PI du client FTP doit être dans l'état Connected
	 * Si ce prérequis n'est pas vérifié, la méthode retourne juste l'état courant 
	 * Si la procédure se passe mal, la connexion doit être fermée et l'état du 
	 * PI du client FTP mis à NotConnected.
	 * @param user: identifiant de l'utilisateur
	 * @param pass: mot de passe de l'uttilisateur
	 * @return état du PI du client FTP en fin de procédure:
	 * @throws IOException
	 */
	public FtpPIStatus logToServer(String user, String pass) throws IOException {
		String response;
		if (status != FtpPIStatus.Connected) {
			return status;
		}
		
		// la commande USER
		sendStringToServer("USER " + user + "\n");
		response = recvStringFromServer();
		if (!cmdOk(response)) {
			close();
			return FtpPIStatus.NotConnected;
		}
		
		// la commande PASS
		sendStringToServer("PASS " + pass + "\n");
		response = recvStringFromServer();
		if (!cmdOk(response)) {
			close();
			return FtpPIStatus.NotConnected;
		}
		
		// tout est ok
		status = FtpPIStatus.LoginOk;
		return status;	
	}
	

	/**
	 * Envoie une ligne de commande CWD (Change Working Directory) au PI du serveur FTP.
	 * @param directory: répertoire destination du CWD
	 * @return true ou false suivant le code du message de retour du PI du serveur
	 * @throws IOException
	 */
	public boolean cwdCmd(String directory) throws IOException {
		String response;
		
		// la commande CWD
		sendStringToServer("CWD " + directory + "\n");
		response = recvStringFromServer();
		return cmdOk(response);
	
	}
	
	
	/**
	 * Envoie une commande RETR au PI du serveur FTP pour demander le télécharment du
	 * fichier filename. Le client FTP sauvera en local ce fichier dans le répetoire courant 
	 * avec le même nom.
	 * @param fileName: nom (relatif au répertoire actuel du serveur FTP) du fichier à télécharger
	 * @return true si un téléchargement (même partiel) a été effectué, false sinon.
	 * @throws IOException
	 */
	public boolean retrCmd(String fileName) throws IOException {
		String response;
		
		
		
		// la commande PASV
		sendStringToServer("PASV\n");
		response = recvStringFromServer();
		if (!cmdOk(response)) {
			return false;
		}
		
		// analyse de la réponse
		PasvResponse pasv = new PasvResponse(response);
		if (!pasv.isOk()) {
			System.err.println("Echec de la commande PASV");
			return false;
		}
		//System.out.println("	Ip address  : " + pasv.getInetAddress());
		//System.out.println("	Port number : " + pasv.getPort());
		
		// connexion au flux ouvert par le serveur
		FtpDataTransferProcess ftpDTP = new FtpDataTransferProcess(pasv.getSocketAddress());
		
		// la commande RETR
		sendStringToServer("RETR " + fileName + "\n");
		response = recvStringFromServer();
		if (!cmdOk(response)) {
			ftpDTP.close();
			return false;
		}

		// reception du fichier
		FileOutputStream fout = new FileOutputStream(fileName);
		ftpDTP.receiveData(fout);
		response = recvStringFromServer();
		if (!cmdOk(response)) {
			System.err.println("Anomalie :" + response);
		}
		fout.close();
		ftpDTP.close();
		return true;
	}



	@Override
	public void run() {
		try {
			
			connectTo(servHostname, port);
			if (status != FtpPIStatus.Connected) {
				System.err.println("Échec de la connexion à " + servHostname + ":" + port);
				return;
			}
			
			if (logToServer(user, pass) != FtpPIStatus.LoginOk) {
				System.err.println("Échec de l'identification");
				return ;
			}
			File file = new File(absoluteFileName);
			String fileName = file.getName();
			//String directory = file.getParent();
			String directory = null;
			if (absoluteFileName.charAt(0) == '/') {
				directory = absoluteFileName.substring(0, absoluteFileName.lastIndexOf('/'));
			}
			
			if ((directory != null) && (!cwdCmd(directory))) {
				System.err.println("Répertoire " + directory + " non trouvé sur le serveur ftp");
				return;
			}
			
			if (!retrCmd(fileName)) {
				System.err.println("Le fichier " + absoluteFileName + " n'a pas pu être téléchargé");
				return ;
			}
		} catch (IOException e) {
			e.printStackTrace();
		}
		
	}
	

}
